unit Unit1;

{$mode objfpc}{$H+}
// {$define hidePanelFS}

interface

uses
  Classes, SysUtils, Forms, Controls, Graphics, Dialogs,
  LCLIntf, LCLType, ExtCtrls, Menus, ComCtrls,
  {$ifndef VER3_0} OpenSSLSockets, {$endif}
  CastleDownload, CastleParameters, CastleClassUtils, CastleImages,
  CastleControl, CastleSceneCore, CastleVectors, CastleScene, CastleViewport,
  CastleNotifications, CastleTransform, CastleCameras, CastleApplicationProperties,
  // X3DNodes, X3DLoad, CastleUIControls, CastleLCLUtils, CastleLCLUtils,
  CastleRectangles, CastleKeysMouse, CastleLog, CastleTimeUtils;

type

  { TForm1 }

  TForm1 = class(TForm)
    CastleControlBase1: TCastleControlBase;
    Panel1: TPanel;
    procedure CastleControlBase1BeforeRender(Sender: TObject);
    procedure CastleControlBase1Press(Sender: TObject; const Event: TInputPressRelease);
    procedure CastleControlBase1Release(Sender: TObject;
      const Event: TInputPressRelease);
    procedure CastleControlBase1Update(Sender: TObject);
    procedure FormCreate(Sender: TObject);
    procedure FormShow(Sender: TObject);
    procedure ToggleFullScreen;
  private
    FullScreen: Boolean;
    Viewport: TCastleViewport;
    Table: TCastleScene;
    Cross: TCastleScene;
    Cube: TCastleScene;
    crossy: Integer;
    stateNotifications: TCastleNotifications;
    infoNotifications: TCastleNotifications;
    screenRect: TFloatRectangle;
    procedure LoadScene(filename: String);
    procedure InfoPanel;
  public

  end;

var
  Form1: TForm1;
  mkShiftState: TModifierKeys = [];

const
  BorderNone = Controls.bsNone;

implementation

{$R *.lfm}

{ TForm1 }

function PositionToSphereAroundCamera(const Viewport: TCastleViewport;
  const Position: TVector2;
  const ScreenCoordinates: Boolean;
  const Radius: Single): TVector3;
var
  R: TFloatRectangle;
  ScreenPosition: TVector2;
  RayOrigin, RayDirection: TVector3;
begin
  R := Viewport.RenderRect;

  if ScreenCoordinates then
    ScreenPosition := Position
  else
    ScreenPosition := Position * Viewport.UIScale + R.LeftBottom;

  Viewport.Camera.CustomRay(R, ScreenPosition, Viewport.Projection, RayOrigin, RayDirection);

  // we know that RayDirection is normalized now
  Result := RayOrigin + RayDirection * Radius;
end;

procedure TForm1.InfoPanel;
var
  msg: String;
begin
  screenRect := Viewport.RenderRect;
  if screenRect.IsEmpty then
    msg := 'ScreenRect = Undefined'
  else
    msg := 'ScreenRect = MinX : ' + FloatToStr(screenRect.Left) + ', ' +
           'MinY : ' + FloatToStr(screenRect.Bottom) + ', ' +
           'MidX : ' + FloatToStr(round(screenRect.Width / 2)) + ', ' +
           'MidY : ' + FloatToStr(round(screenRect.Height / 2)) + ', ' +
           'Width : ' + FloatToStr(screenRect.Width) + ', ' +
           'Height : ' + FloatToStr(screenRect.Height) + ', ' +
           'FPS : ' + FormatFloat('##0.00', CastleControlBase1.Fps.RealFps);
  Panel1.Caption := msg;

  WritelnLog(msg);
end;

procedure TForm1.ToggleFullScreen;
begin
  FullScreen := not FullScreen;

  if FullScreen then
    begin
      {$ifdef unix}
      ShowWindow(Handle, SW_SHOWFULLSCREEN);
      {$else}
      BorderStyle := BorderNone;
      WindowState := wsFullScreen;
      FormStyle := fsSystemStayOnTop;
      {$endif}
      {$ifdef hidePanelFS}
      Panel1.Visible := False;
      {$endif}
    end
  else
    begin
      {$ifdef unix}
      ShowWindow(Handle, SW_SHOWNORMAL);
      {$else}
      WindowState := wsNormal; // wsMinimized, wsMaximized, wsFullScreen
      BorderStyle := bsSizeable; // bsDialog, bsNone, bsSingle
      FormStyle := fsNormal;
      ShowInTaskBar := stDefault; // stAlways, stNever
      {$endif}
      {$ifdef hidePanelFS}
      Panel1.Visible := True;
      {$endif}
    end;

    InfoPanel;
end;

procedure TForm1.CastleControlBase1Press(Sender: TObject;
  const Event: TInputPressRelease);
var
  TopMostScene: TCastleTransform;
begin
  if Event.IsKey(keyF11) then
    begin
      ToggleFullScreen;
    end;

  if Event.IsKey(keyR) then
    begin
      Viewport.Camera.GoToInitial;
      if Viewport.Navigation is TCastleExamineNavigation then
        (Viewport.Navigation as TCastleExamineNavigation).StopRotating;
    end;

  if Event.IsKey(keyX) then
    begin
      inc(crossy);
      if crossy > 1 then
        crossy := -1;
    end;

  if not (stateNotifications = nil) then
  begin
    if Event.IsMouseButton(mbLeft) then
    begin
      if Viewport.MouseRayHit <> nil then
        TopMostScene := Viewport.MouseRayHit.First.Item
      else
        TopMostScene := nil;

      if TopMostScene <> nil then
        infoNotifications.Show('Clicked on scene ' + IntToStr(TopMostScene.Tag))
      else
        infoNotifications.Show('Clicked on empty space')
    end;
  end;

  mkShiftState := Event.ModifiersDown;

end;

procedure TForm1.CastleControlBase1Release(Sender: TObject;
  const Event: TInputPressRelease);
begin
  mkShiftState := Event.ModifiersDown;
end;

procedure TForm1.CastleControlBase1BeforeRender(Sender: TObject);
const
  SecsPerRot = 4;
var
  msg: String;
  RotDir, Pos, Dir, Up, CrossProjection, CubeProjection: TVector3;
  theta: Single;
begin
  Viewport.Camera.GetView(Pos, Dir, Up);
  theta := ((CastleGetTickCount64 mod (SecsPerRot * 1000)) / (SecsPerRot * 1000)) * (Pi * 2);

  // Horrible Vector2 bodge to action X key - sorry
  CrossProjection := PositionToSphereAroundCamera(Viewport, Vector2(256, (screenRect.Height - round(screenRect.Height / 2) + (round(screenRect.Height / 4) * crossy))), True, 200);
  RotDir := RotatePointAroundAxisRad(theta, Dir, Up);
  Cross.Exists := true;
  Cross.SetView(RotDir, Up);
  Cross.Translation := CrossProjection;

  CubeProjection := PositionToSphereAroundCamera(Viewport, Vector2(screenRect.Width - 256, screenRect.Height - 256), True, 200);
  RotDir := RotatePointAroundAxisRad(-theta, Dir, Up);
  Cube.Exists := true;
  Cube.SetView(RotDir, Up);
  Cube.Translation := CubeProjection;

  if (mkShift in mkShiftState) and (not (mkCtrl in mkShiftState)) then
    begin
      msg := 'Cross' + LineEnding +
           'Pos = X : ' + FormatFloat('####0.00', Cross.Translation.X) +
           ', Y : ' + FormatFloat('####0.00', Cross.Translation.Y) +
           ', Z : ' + FormatFloat('####0.00', Cross.Translation.Z) + LineEnding +
           'Dir = X : ' + FormatFloat('####0.00', Cross.Direction.X) +
           ', Y : ' + FormatFloat('####0.00', Cross.Direction.Y) +
           ', Z : ' + FormatFloat('####0.00', Cross.Direction.Z) + LineEnding +
           'Up = X : ' + FormatFloat('####0.00', Cross.Up.X) +
           ', Y : ' + FormatFloat('####0.00', Cross.Up.Y) +
           ', Z : ' + FormatFloat('####0.00', Cross.Up.Z);
    end
  else if (not (mkShift in mkShiftState)) and (mkCtrl in mkShiftState) then
    begin
      msg := 'Cube' + LineEnding +
           'Pos = X : ' + FormatFloat('####0.00', Cube.Translation.X) +
           ', Y : ' + FormatFloat('####0.00', Cube.Translation.Y) +
           ', Z : ' + FormatFloat('####0.00', Cube.Translation.Z) + LineEnding +
           'Dir = X : ' + FormatFloat('####0.00', Cube.Direction.X) +
           ', Y : ' + FormatFloat('####0.00', Cube.Direction.Y) +
           ', Z : ' + FormatFloat('####0.00', Cube.Direction.Z) + LineEnding +
           'Up = X : ' + FormatFloat('####0.00', Cube.Up.X) +
           ', Y : ' + FormatFloat('####0.00', Cube.Up.Y) +
           ', Z : ' + FormatFloat('####0.00', Cube.Up.Z);
    end
  else if (mkShift in mkShiftState) and (mkCtrl in mkShiftState) then
    begin
      msg := 'Table' + LineEnding +
           'Pos = X : ' + FormatFloat('####0.00', Table.Translation.X) +
           ', Y : ' + FormatFloat('####0.00', Table.Translation.Y) +
           ', Z : ' + FormatFloat('####0.00', Table.Translation.Z) + LineEnding +
           'Dir = X : ' + FormatFloat('####0.00', Table.Direction.X) +
           ', Y : ' + FormatFloat('####0.00', Table.Direction.Y) +
           ', Z : ' + FormatFloat('####0.00', Table.Direction.Z) + LineEnding +
           'Up = X : ' + FormatFloat('####0.00', Table.Up.X) +
           ', Y : ' + FormatFloat('####0.00', Table.Up.Y) +
           ', Z : ' + FormatFloat('####0.00', Table.Up.Z);
    end
  else
    begin
       msg := 'Camera' + LineEnding +
           'Pos = X : ' + FormatFloat('####0.00', Pos.X) +
           ', Y : ' + FormatFloat('####0.00', Pos.Y) +
           ', Z : ' + FormatFloat('####0.00', Pos.Z) + LineEnding +
           'Dir = X : ' + FormatFloat('####0.00', Dir.X) +
           ', Y : ' + FormatFloat('####0.00', Dir.Y) +
           ', Z : ' + FormatFloat('####0.00', Dir.Z) + LineEnding +
           'Up = X : ' + FormatFloat('####0.00', Up.X) +
           ', Y : ' + FormatFloat('####0.00', Up.Y) +
           ', Z : ' + FormatFloat('####0.00', Up.Z);
    end;
  stateNotifications.Show(msg);

end;

procedure TForm1.CastleControlBase1Update(Sender: TObject);
begin
  InfoPanel;
end;

procedure TForm1.FormShow(Sender: TObject);
begin
  Application.ProcessMessages;
end;

procedure TForm1.LoadScene(filename: String);
begin
  if not (Table = nil) then
    begin
      Table.Free;
      Table := nil;
    end;

  ApplicationProperties.LimitFPS := 25;
  CastleControlBase1.Enabled := False;

  Viewport := TCastleViewport.Create(Application);
  Viewport.FullSize := true;
  Viewport.AutoCamera := true; // instead of this, you could do "Viewport.Camera.SetView(...)"
  Viewport.AutoNavigation := true; // instead of this, you could do "Viewport.Navigation := ..."

  // Shows various state messages
  stateNotifications := TCastleNotifications.Create(Application);
  // stateNotifications.Font := 'Courier New'; // Set font face to anything mono-spaced
  stateNotifications.MaxMessages := 4;
  stateNotifications.Anchor(hpLeft, 10);
  stateNotifications.Anchor(vpBottom, 10);

  // Interaction messages
  infoNotifications := TCastleNotifications.Create(Application);
  infoNotifications.MaxMessages := 4;
  infoNotifications.Anchor(hpRight, -10);
  infoNotifications.Anchor(vpBottom, 10);

  CastleControlBase1.Controls.InsertFront(Viewport);
  CastleControlBase1.Controls.InsertFront(stateNotifications);
  CastleControlBase1.Controls.InsertFront(infoNotifications);

  Table := TCastleScene.Create(Application);
  Table.Spatial := [ssRendering, ssDynamicCollisions];
  Table.ProcessEvents := true;

  Table.Load(filename);
  Table.Tag := 1;
  Viewport.Items.Add(Table);
  
  Cross := TCastleScene.Create(Application);
  Cross.load('castle-data:/cross3.x3dv');
  Cross.Scale := Vector3(1, 1, 1);
  Cross.Tag := 2;
  Cross.Exists := false;
  Viewport.Items.Add(Cross);

  Cube := TCastleScene.Create(Application);
  Cube.PrimitiveGeometry := pgBox;
  Cube.Scale := Vector3(10, 10, 10);
  Cube.Tag := 3;
  Cube.Exists := false;
  Viewport.Items.Add(Cube);

  Viewport.Items.MainScene := Table;

  CastleControlBase1.Enabled:=True;

  InfoPanel;
end;

procedure TForm1.FormCreate(Sender: TObject);
begin
  FullScreen := False;
  Form1.Width := 1280;
  Form1.Height := 800;
  Form1.Caption := 'view3d';
  crossy := -1;
  InitializeLog;
  LoadScene('castle-data:/table.x3dv');
end;

end.

